@@ -11,17 +11,6 @@
 
 #include <QCollator>
 
-void KCompletionPrivate::init()
-{
-    completionMode = KCompletion::CompletionPopup;
-    treeNodeAllocator = KCompTreeNode::allocator(); // keep strong-ref to allocator instance
-    treeRoot = new KCompTreeNode;
-    beep = true;
-    ignoreCase = false;
-    hasMultipleMatches = false;
-    rotationIndex = 0;
-}
-
 void KCompletionPrivate::addWeightedItem(const QString &item)
 {
     Q_Q(KCompletion);
@@ -37,7 +26,11 @@ void KCompletionPrivate::addWeightedItem(const QString &item)
     int index = item.lastIndexOf(QLatin1Char(':'));
     if (index > 0) {
         bool ok;
+#if QT_VERSION >= QT_VERSION_CHECK(6, 0, 0)
+        weight = QStringView(item).mid(index + 1).toUInt(&ok);
+#else
         weight = item.midRef(index + 1).toUInt(&ok);
+#endif
         if (!ok) {
             weight = 0;
         }
@@ -52,19 +45,17 @@ void KCompletionPrivate::addWeightedItem(const QString &item)
 // tries to complete "string" from the tree-root
 QString KCompletionPrivate::findCompletion(const QString &string)
 {
-    QChar ch;
     QString completion;
-    const KCompTreeNode *node = treeRoot;
+    const KCompTreeNode *node = m_treeRoot.get();
 
     // start at the tree-root and try to find the search-string
-    for (int i = 0; i < string.length(); i++) {
-        ch = string.at(i);
+    for (const auto ch : string) {
         node = node->find(ch);
 
         if (node) {
             completion += ch;
         } else {
-            return QString();    // no completion
+            return QString(); // no completion
         }
     }
 
@@ -125,11 +116,16 @@ QString KCompletionPrivate::findCompletion(const QString &string)
     return completion;
 }
 
+void KCompletionPrivate::defaultSort(QStringList &stringList)
+{
+    QCollator c;
+    c.setCaseSensitivity(Qt::CaseSensitive);
+    std::stable_sort(stringList.begin(), stringList.end(), c);
+}
+
 KCompletion::KCompletion()
     : d_ptr(new KCompletionPrivate(this))
 {
-    Q_D(KCompletion);
-    d->init();
     setOrder(Insertion);
 }
 
@@ -171,15 +167,11 @@ void KCompletion::setItems(const QStringList &itemList)
 void KCompletion::insertItems(const QStringList &items)
 {
     Q_D(KCompletion);
-    bool weighted = (d->order == Weighted);
-    QStringList::ConstIterator it;
-    if (weighted) {   // determine weight
-        for (it = items.begin(); it != items.end(); ++it) {
-            d->addWeightedItem(*it);
-        }
-    } else {
-        for (it = items.begin(); it != items.end(); ++it) {
-            addItem(*it, 0);
+    for (const auto &str : items) {
+        if (d->order == Weighted) {
+            d->addWeightedItem(str);
+        } else {
+            addItem(str, 0);
         }
     }
 }
@@ -187,17 +179,15 @@ void KCompletion::insertItems(const QStringList &items)
 QStringList KCompletion::items() const
 {
     Q_D(const KCompletion);
-    KCompletionMatchesWrapper list; // unsorted
-    bool addWeight = (d->order == Weighted);
-    list.extractStringsFromNode(d->treeRoot, QString(), addWeight);
-
+    KCompletionMatchesWrapper list(d->sorterFunction); // unsorted
+    list.extractStringsFromNode(d->m_treeRoot.get(), QString(), d->order == Weighted);
     return list.list();
 }
 
 bool KCompletion::isEmpty() const
 {
     Q_D(const KCompletion);
-    return (d->treeRoot->childrenCount() == 0);
+    return (d->m_treeRoot->childrenCount() == 0);
 }
 
 void KCompletion::postProcessMatch(QString *) const
@@ -229,7 +219,7 @@ void KCompletion::addItem(const QString &item, uint weight)
         return;
     }
 
-    KCompTreeNode *node = d->treeRoot;
+    KCompTreeNode *node = d->m_treeRoot.get();
     int len = item.length();
 
     bool sorted = (d->order == Sorted);
@@ -241,16 +231,16 @@ void KCompletion::addItem(const QString &item, uint weight)
     for (int i = 0; i < len; i++) {
         node = node->insert(item.at(i), sorted);
         if (weighted) {
-            node->confirm(weight - 1);    // node->insert() sets weighting to 1
+            node->confirm(weight - 1); // node->insert() sets weighting to 1
         }
     }
 
     // add 0x0-item as delimiter with evtl. weight
-    node = node->insert(0x0, true);
+    node = node->insert(QChar(0x0), true);
     if (weighted) {
         node->confirm(weight - 1);
     }
-//     qDebug("*** added: %s (%i)", item.toLatin1().constData(), node->weight());
+    //     qDebug("*** added: %s (%i)", item.toLatin1().constData(), node->weight());
 }
 
 void KCompletion::removeItem(const QString &item)
@@ -260,7 +250,7 @@ void KCompletion::removeItem(const QString &item)
     d->rotationIndex = 0;
     d->lastString.clear();
 
-    d->treeRoot->remove(item);
+    d->m_treeRoot->remove(item);
 }
 
 void KCompletion::clear()
@@ -270,8 +260,7 @@ void KCompletion::clear()
     d->rotationIndex = 0;
     d->lastString.clear();
 
-    delete d->treeRoot;
-    d->treeRoot = new KCompTreeNode;
+    d->m_treeRoot.reset(new KCompTreeNode);
 }
 
 QString KCompletion::makeCompletion(const QString &string)
@@ -281,7 +270,7 @@ QString KCompletion::makeCompletion(const QString &string)
         return QString();
     }
 
-    //qDebug() << "KCompletion: completing: " << string;
+    // qDebug() << "KCompletion: completing: " << string;
 
     d->matches.clear();
     d->rotationIndex = 0;
@@ -290,25 +279,23 @@ QString KCompletion::makeCompletion(const QString &string)
 
     // in Shell-completion-mode, emit all matches when we get the same
     // complete-string twice
-    if (d->completionMode == CompletionShell &&
-            string == d->lastString) {
+    if (d->completionMode == CompletionShell && string == d->lastString) {
         // Don't use d->matches since calling postProcessMatches()
         // on d->matches here would interfere with call to
         // postProcessMatch() during rotation
 
-        d->matches.findAllCompletions(d->treeRoot, string, d->ignoreCase, d->hasMultipleMatches);
+        d->matches.findAllCompletions(d->m_treeRoot.get(), string, d->ignoreCase, d->hasMultipleMatches);
         QStringList l = d->matches.list();
         postProcessMatches(&l);
-        emit matches(l);
+        Q_EMIT matches(l);
 
         return QString();
     }
 
     QString completion;
     // in case-insensitive popup mode, we search all completions at once
-    if (d->completionMode == CompletionPopup ||
-            d->completionMode == CompletionPopupAuto) {
-        d->matches.findAllCompletions(d->treeRoot, string, d->ignoreCase, d->hasMultipleMatches);
+    if (d->completionMode == CompletionPopup || d->completionMode == CompletionPopupAuto) {
+        d->matches.findAllCompletions(d->m_treeRoot.get(), string, d->ignoreCase, d->hasMultipleMatches);
         if (!d->matches.isEmpty()) {
             completion = d->matches.first();
         }
@@ -317,7 +304,7 @@ QString KCompletion::makeCompletion(const QString &string)
     }
 
     if (d->hasMultipleMatches) {
-        emit multipleMatches();
+        Q_EMIT multipleMatches();
     }
 
     d->lastString = string;
@@ -325,9 +312,9 @@ QString KCompletion::makeCompletion(const QString &string)
 
     postProcessMatch(&completion);
 
-    if (!string.isEmpty()) {   // only emit match when string is not empty
-        //qDebug() << "KCompletion: Match: " << completion;
-        emit match(completion);
+    if (!string.isEmpty()) { // only emit match when string is not empty
+        // qDebug() << "KCompletion: Match: " << completion;
+        Q_EMIT match(completion);
     }
 
     return completion;
@@ -337,8 +324,8 @@ QStringList KCompletion::substringCompletion(const QString &string) const
 {
     Q_D(const KCompletion);
     // get all items in the tree, eventually in sorted order
-    KCompletionMatchesWrapper allItems(d->order);
-    allItems.extractStringsFromNode(d->treeRoot, QString(), false);
+    KCompletionMatchesWrapper allItems(d->sorterFunction, d->order);
+    allItems.extractStringsFromNode(d->m_treeRoot.get(), QString(), false);
 
     QStringList list = allItems.list();
 
@@ -347,21 +334,15 @@ QStringList KCompletion::substringCompletion(const QString &string) const
         return list;
     }
 
-    if (string.isEmpty()) {   // shortcut
-        postProcessMatches(&list);
-        return list;
+    if (!string.isEmpty()) { // If it's empty, nothing to compare
+        auto it = std::remove_if(list.begin(), list.end(), [&string](const QString &item) {
+            return !item.contains(string, Qt::CaseInsensitive); // always case insensitive
+        });
+        list.erase(it, list.end());
     }
 
-    QStringList matches;
-    for (QStringList::ConstIterator it = list.constBegin(), total = list.constEnd(); it != total; ++it) {
-        QString item = *it;
-        if (item.indexOf(string, 0, Qt::CaseInsensitive) != -1) {     // always case insensitive
-            postProcessMatch(&item);
-            matches.append(item);
-        }
-    }
-
-    return matches;
+    postProcessMatches(&list);
+    return list;
 }
 
 void KCompletion::setCompletionMode(CompletionMode mode)
@@ -376,15 +357,33 @@ KCompletion::CompletionMode KCompletion::completionMode() const
     return d->completionMode;
 }
 
+void KCompletion::setShouldAutoSuggest(const bool shouldAutoSuggest)
+{
+    Q_D(KCompletion);
+    d->shouldAutoSuggest = shouldAutoSuggest;
+}
+
+bool KCompletion::shouldAutoSuggest() const
+{
+    Q_D(const KCompletion);
+    return d->shouldAutoSuggest;
+}
+
+void KCompletion::setSorterFunction(SorterFunction sortFunc)
+{
+    Q_D(KCompletion);
+    d->sorterFunction = sortFunc ? sortFunc : KCompletionPrivate::defaultSort;
+}
+
 QStringList KCompletion::allMatches()
 {
     Q_D(KCompletion);
     // Don't use d->matches since calling postProcessMatches()
     // on d->matches here would interfere with call to
     // postProcessMatch() during rotation
-    KCompletionMatchesWrapper matches(d->order);
+    KCompletionMatchesWrapper matches(d->sorterFunction, d->order);
     bool dummy;
-    matches.findAllCompletions(d->treeRoot, d->lastString, d->ignoreCase, dummy);
+    matches.findAllCompletions(d->m_treeRoot.get(), d->lastString, d->ignoreCase, dummy);
     QStringList l = matches.list();
     postProcessMatches(&l);
     return l;
@@ -396,9 +395,9 @@ KCompletionMatches KCompletion::allWeightedMatches()
     // Don't use d->matches since calling postProcessMatches()
     // on d->matches here would interfere with call to
     // postProcessMatch() during rotation
-    KCompletionMatchesWrapper matches(d->order);
+    KCompletionMatchesWrapper matches(d->sorterFunction, d->order);
     bool dummy;
-    matches.findAllCompletions(d->treeRoot, d->lastString, d->ignoreCase, dummy);
+    matches.findAllCompletions(d->m_treeRoot.get(), d->lastString, d->ignoreCase, dummy);
     KCompletionMatches ret(matches);
     postProcessMatches(&ret);
     return ret;
@@ -407,9 +406,9 @@ KCompletionMatches KCompletion::allWeightedMatches()
 QStringList KCompletion::allMatches(const QString &string)
 {
     Q_D(KCompletion);
-    KCompletionMatchesWrapper matches(d->order);
+    KCompletionMatchesWrapper matches(d->sorterFunction, d->order);
     bool dummy;
-    matches.findAllCompletions(d->treeRoot, string, d->ignoreCase, dummy);
+    matches.findAllCompletions(d->m_treeRoot.get(), string, d->ignoreCase, dummy);
     QStringList l = matches.list();
     postProcessMatches(&l);
     return l;
@@ -418,9 +417,9 @@ QStringList KCompletion::allMatches(const QString &string)
 KCompletionMatches KCompletion::allWeightedMatches(const QString &string)
 {
     Q_D(KCompletion);
-    KCompletionMatchesWrapper matches(d->order);
+    KCompletionMatchesWrapper matches(d->sorterFunction, d->order);
     bool dummy;
-    matches.findAllCompletions(d->treeRoot, string, d->ignoreCase, dummy);
+    matches.findAllCompletions(d->m_treeRoot.get(), string, d->ignoreCase, dummy);
     KCompletionMatches ret(matches);
     postProcessMatches(&ret);
     return ret;
@@ -454,28 +453,28 @@ QString KCompletion::nextMatch()
     d->lastMatch = d->currentMatch;
 
     if (d->matches.isEmpty()) {
-        d->matches.findAllCompletions(d->treeRoot, d->lastString, d->ignoreCase, d->hasMultipleMatches);
+        d->matches.findAllCompletions(d->m_treeRoot.get(), d->lastString, d->ignoreCase, d->hasMultipleMatches);
         if (!d->matches.isEmpty()) {
             completion = d->matches.first();
         }
         d->currentMatch = completion;
         d->rotationIndex = 0;
         postProcessMatch(&completion);
-        emit match(completion);
+        Q_EMIT match(completion);
         return completion;
     }
 
     QStringList matches = d->matches.list();
-    d->lastMatch = matches[ d->rotationIndex++ ];
+    d->lastMatch = matches[d->rotationIndex++];
 
     if (d->rotationIndex == matches.count()) {
         d->rotationIndex = 0;
     }
 
-    completion = matches[ d->rotationIndex ];
+    completion = matches[d->rotationIndex];
     d->currentMatch = completion;
     postProcessMatch(&completion);
-    emit match(completion);
+    Q_EMIT match(completion);
     return completion;
 }
 
@@ -492,19 +491,19 @@ QString KCompletion::previousMatch()
     d->lastMatch = d->currentMatch;
 
     if (d->matches.isEmpty()) {
-        d->matches.findAllCompletions(d->treeRoot, d->lastString, d->ignoreCase, d->hasMultipleMatches);
+        d->matches.findAllCompletions(d->m_treeRoot.get(), d->lastString, d->ignoreCase, d->hasMultipleMatches);
         if (!d->matches.isEmpty()) {
             completion = d->matches.last();
         }
         d->currentMatch = completion;
         d->rotationIndex = 0;
         postProcessMatch(&completion);
-        emit match(completion);
+        Q_EMIT match(completion);
         return completion;
     }
 
     QStringList matches = d->matches.list();
-    d->lastMatch = matches[ d->rotationIndex ];
+    d->lastMatch = matches[d->rotationIndex];
 
     if (d->rotationIndex == 0) {
         d->rotationIndex = matches.count();
@@ -512,334 +511,13 @@ QString KCompletion::previousMatch()
 
     d->rotationIndex--;
 
-    completion = matches[ d->rotationIndex ];
+    completion = matches[d->rotationIndex];
     d->currentMatch = completion;
     postProcessMatch(&completion);
-    emit match(completion);
+    Q_EMIT match(completion);
     return completion;
 }
 
-/////////////////////////////////
-/////////
-
-// Implements the tree. Every node is a QChar and has a list of children, which
-// are Nodes as well.
-// QChar( 0x0 ) is used as the delimiter of a string; the last child of each
-// inserted string is 0x0.
-
-KCompTreeNode::~KCompTreeNode()
-{
-    // delete all children
-    KCompTreeNode *cur = m_children.begin();
-    while (cur) {
-        KCompTreeNode *next = cur->m_next;
-        delete m_children.remove(cur);
-        cur = next;
-    }
-}
-
-// Adds a child-node "ch" to this node. If such a node is already existent,
-// it will not be created. Returns the new/existing node.
-KCompTreeNode *KCompTreeNode::insert(const QChar &ch, bool sorted)
-{
-    KCompTreeNode *child = find(ch);
-    if (!child) {
-        child = new KCompTreeNode(ch);
-
-        // FIXME, first (slow) sorted insertion implementation
-        if (sorted) {
-            KCompTreeNode *prev = nullptr;
-            KCompTreeNode *cur = m_children.begin();
-            while (cur) {
-                if (ch > *cur) {
-                    prev = cur;
-                    cur = cur->m_next;
-                } else {
-                    break;
-                }
-            }
-            if (prev) {
-                m_children.insert(prev, child);
-            } else {
-                m_children.prepend(child);
-            }
-        }
-
-        else {
-            m_children.append(child);
-        }
-    }
-
-    // implicit weighting: the more often an item is inserted, the higher
-    // priority it gets.
-    child->confirm();
-
-    return child;
-}
-
-// Iteratively removes a string from the tree. The nicer recursive
-// version apparently was a little memory hungry (see #56757)
-void KCompTreeNode::remove(const QString &str)
-{
-    QString string = str;
-    string += QChar(0x0);
-
-    QVector<KCompTreeNode *> deletables(string.length() + 1);
-
-    KCompTreeNode *child = nullptr;
-    KCompTreeNode *parent = this;
-    deletables.replace(0, parent);
-
-    int i = 0;
-    for (; i < string.length(); i++) {
-        child = parent->find(string.at(i));
-        if (child) {
-            deletables.replace(i + 1, child);
-        } else {
-            break;
-        }
-
-        parent = child;
-    }
-
-    for (; i >= 1; i--) {
-        parent = deletables.at(i - 1);
-        child = deletables.at(i);
-        if (child->m_children.count() == 0) {
-            delete parent->m_children.remove(child);
-        }
-    }
-}
-
-QStringList KCompletionMatchesWrapper::list() const
-{
-    if (m_sortedList && m_dirty) {
-        m_sortedList->sort();
-        m_dirty = false;
-
-        m_stringList.clear();
-
-        // high weight == sorted last -> reverse the sorting here
-        QList<KSortableItem<QString> >::const_iterator it;
-        for (it = m_sortedList->constBegin(); it != m_sortedList->constEnd(); ++it) {
-            m_stringList.prepend((*it).value());
-        }
-    } else if (m_compOrder == KCompletion::Sorted) {
-        QCollator c;
-        c.setCaseSensitivity(Qt::CaseSensitive);
-        std::stable_sort(m_stringList.begin(), m_stringList.end(), c);
-    }
-
-    return m_stringList;
-}
-
-void KCompletionMatchesWrapper::findAllCompletions(const KCompTreeNode *treeRoot,const QString &string, bool ignoreCase, bool &hasMultipleMatches)
-{
-    //qDebug() << "*** finding all completions for " << string;
-
-    if (string.isEmpty()) {
-        return;
-    }
-
-    if (ignoreCase) {   // case insensitive completion
-        extractStringsFromNodeCI(treeRoot, QString(), string);
-        hasMultipleMatches = (count() > 1);
-        return;
-    }
-
-    QChar ch;
-    QString completion;
-    const KCompTreeNode *node = treeRoot;
-
-    // start at the tree-root and try to find the search-string
-    for (int i = 0; i < string.length(); i++) {
-        ch = string.at(i);
-        node = node->find(ch);
-
-        if (node) {
-            completion += ch;
-        } else {
-            return;    // no completion -> return empty list
-        }
-    }
-
-    // Now we have the last node of the to be completed string.
-    // Follow it as long as it has exactly one child (= longest possible
-    // completion)
-
-    while (node->childrenCount() == 1) {
-        node = node->firstChild();
-        if (!node->isNull()) {
-            completion += *node;
-        }
-        // qDebug() << completion << node->latin1();
-    }
-
-    // there is just one single match)
-    if (node->childrenCount() == 0) {
-        append(node->weight(), completion);
-    }
-
-    else {
-        // node has more than one child
-        // -> recursively find all remaining completions
-        hasMultipleMatches = true;
-        extractStringsFromNode(node, completion);
-    }
-}
-
-void KCompletionMatchesWrapper::extractStringsFromNode(const KCompTreeNode *node, const QString &beginning, bool addWeight)
-{
-    if (!node) {
-        return;
-    }
-
-    // qDebug() << "Beginning: " << beginning;
-    const KCompTreeChildren *list = node->children();
-    QString string;
-    QString w;
-
-    // loop thru all children
-    for (KCompTreeNode *cur = list->begin(); cur; cur = cur->m_next) {
-        string = beginning;
-        node = cur;
-        if (!node->isNull()) {
-            string += *node;
-        }
-
-        while (node && node->childrenCount() == 1) {
-            node = node->firstChild();
-            if (node->isNull()) {
-                break;
-            }
-            string += *node;
-        }
-
-        if (node && node->isNull()) {   // we found a leaf
-            if (addWeight) {
-                // add ":num" to the string to store the weighting
-                string += QLatin1Char(':');
-                w.setNum(node->weight());
-                string.append(w);
-            }
-            append(node->weight(), string);
-        }
-
-        // recursively find all other strings.
-        if (node && node->childrenCount() > 1) {
-            extractStringsFromNode(node, string, addWeight);
-        }
-    }
-}
-
-void KCompletionMatchesWrapper::extractStringsFromNodeCI(const KCompTreeNode *node, const QString &beginning, const QString &restString)
-{
-    if (restString.isEmpty()) {
-        extractStringsFromNode(node, beginning, false /*noweight*/);
-        return;
-    }
-
-    QChar ch1 = restString.at(0);
-    QString newRest = restString.mid(1);
-    KCompTreeNode *child1, *child2;
-
-    child1 = node->find(ch1);   // the correct match
-    if (child1)
-        extractStringsFromNodeCI(child1, beginning + QChar(*child1), newRest);
-
-    // append the case insensitive matches, if available
-    if (ch1.isLetter()) {
-        // find out if we have to lower or upper it. Is there a better way?
-        QChar ch2 = ch1.toLower();
-        if (ch1 == ch2) {
-            ch2 = ch1.toUpper();
-        }
-        if (ch1 != ch2) {
-            child2 = node->find(ch2);
-            if (child2)
-                extractStringsFromNodeCI(child2, beginning + QChar(*child2), newRest);
-        }
-    }
-}
-
-void KCompTreeNodeList::append(KCompTreeNode *item)
-{
-    m_count++;
-    if (!m_last) {
-        m_last = item;
-        m_last->m_next = nullptr;
-        m_first = item;
-        return;
-    }
-    m_last->m_next = item;
-    item->m_next = nullptr;
-    m_last = item;
-}
-
-void KCompTreeNodeList::prepend(KCompTreeNode *item)
-{
-    m_count++;
-    if (!m_last) {
-        m_last = item;
-        m_last->m_next = nullptr;
-        m_first = item;
-        return;
-    }
-    item->m_next = m_first;
-    m_first = item;
-}
-
-void KCompTreeNodeList::insert(KCompTreeNode *after, KCompTreeNode *item)
-{
-    if (!after) {
-        append(item);
-        return;
-    }
-
-    m_count++;
-
-    item->m_next = after->m_next;
-    after->m_next = item;
-
-    if (after == m_last) {
-        m_last = item;
-    }
-}
-
-KCompTreeNode *KCompTreeNodeList::remove(KCompTreeNode *item)
-{
-    if (!m_first || !item) {
-        return nullptr;
-    }
-    KCompTreeNode *cur = nullptr;
-
-    if (item == m_first) {
-        m_first = m_first->m_next;
-    } else {
-        cur = m_first;
-        while (cur && cur->m_next != item) {
-            cur = cur->m_next;
-        }
-        if (!cur) {
-            return nullptr;
-        }
-        cur->m_next = item->m_next;
-    }
-    if (item == m_last) {
-        m_last = cur;
-    }
-    m_count--;
-    return item;
-}
-
-KCompTreeNode *KCompTreeNodeList::at(uint index) const
-{
-    KCompTreeNode *cur = m_first;
-    while (index-- && cur) {
-        cur = cur->m_next;
-    }
-    return cur;
-}
-
-QSharedPointer<KZoneAllocator> KCompTreeNode::m_alloc(new KZoneAllocator(8*1024));
+QSharedPointer<KZoneAllocator> KCompTreeNode::m_alloc(new KZoneAllocator(8 * 1024));
 
+#include "moc_kcompletion.cpp"
