@@ -8,11 +8,11 @@
 #ifndef KCOMPLETIONMATCHES_H
 #define KCOMPLETIONMATCHES_H
 
-#include <ksortablelist.h>
 #include <kcompletion_export.h>
+#include <ksortablelist.h>
 
 #include <QStringList>
-#include <QScopedPointer>
+#include <memory>
 
 class KCompletionMatchesWrapper;
 class KCompletionMatchesPrivate;
@@ -89,7 +89,7 @@ public:
     bool sorting() const;
 
 private:
-    const QScopedPointer<KCompletionMatchesPrivate> d_ptr;
+    std::unique_ptr<KCompletionMatchesPrivate> const d_ptr;
 };
 
 #endif // KCOMPLETIONMATCHES_H
